@@ -6,7 +6,7 @@
  *
  *       Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
  *
- *    $Id: hlst.c,v 1.26 2001/01/21 14:18:26 renaud Exp $ 
+ *    $Id: hlst.c,v 1.33 2003/02/27 10:09:57 renaud Exp $ 
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Library General Public
@@ -30,6 +30,10 @@
 #define __HLST_INTERNAL__
 #include "hlst.h"
 
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#endif
+
 /* ------------------------------------------------------------------------- *
  *                      private definitions                                  *
  * ------------------------------------------------------------------------- */
@@ -41,6 +45,9 @@
 /* default number of hash buckets per list */
 #define DEFAULT_BUCKETS 53
 
+/* global factor to be applied internally to any estimated_size_hint */
+#define DEFAULT_PERCENTAGE_COMPRESSOR 80
+
 typedef
 struct _sorter {
   int                  dirty ;
@@ -93,6 +100,19 @@ static const hash_defs hints [] = {
   {0,0}
 };
 
+static unsigned 
+size_hint_percentage_compressor = DEFAULT_PERCENTAGE_COMPRESSOR ;
+
+/* used for custum sorting -- not thread safe */
+static void *sorter_desc ;
+static int (*sorter_fn)(void*,const char*,unsigned,const char*,unsigned);
+
+#ifdef USE_PTHREADS
+/* some global sync for logging etc  -- make custum sorting thread safe */
+static int glob_mutex_initialized = 0;
+static pthread_mutex_t glob_mutex; 
+#endif
+
 /* ------------------------------------------------------------------------- *
  *               hashed index ank key length calculation                     *
  * ------------------------------------------------------------------------- */
@@ -100,29 +120,57 @@ static const hash_defs hints [] = {
 /* make a hash from the argument key string */
 #define _GET_HASH_AND_STRLEN( h, H, L, key) {	\
   const char *s = (key) ;			\
-  (H) = 0 ;					\
-  (L) = 1 ;					\
-  while (*s) {					\
+  (H) = *s ;					\
+  (L) =  1 ;					\
+  goto S; do {			                \
       (H) *= (h)->z.fac ; /* shift accu  */	\
-      (H) +=     * s ++ ; /* get char    */	\
-      (H) %= (h)->z.mod ; /* module size */	\
-      (L) ++            ; /* word length */	}}
+      (H) +=        * s ; /* get char    */	\
+      (L) ++            ; /* word length */	\
+    S:(H) %= (h)->z.mod ; /* module size */	\
+  } while (*s ++) ;				}
 
 /* make a hash from a generic argument key of given length */
 #define _GET_HASH( h, H, len, key) {		\
   const char *s = (key) ;			\
   int l = (len) ;				\
-  (H)   =     0 ;				\
-  while (l --) {				\
+  (H)   =    *s ;				\
+  goto T; do {					\
       (H) *= (h)->z.fac ; /* shift accu  */	\
-      (H) +=     * s ++ ; /* get char    */	\
-      (H) %= (h)->z.mod ; /* module size */	}}
+      (H) +=     * ++ s ; /* get char    */	\
+    T:(H) %= (h)->z.mod ; /* module size */	\
+  } while (-- l);				}
 
 /* combine that methods, above */
 #define GET_HASH_AND_LEN( h, H, len, key)	\
   { if (len) _GET_HASH (h,H,len,key) else _GET_HASH_AND_STRLEN(h,H,len,key) }
 
 
+#ifdef USE_PTHREADS
+# define mutex_init(x)     pthread_mutex_init    (x, 0)
+# define mutex_destroy(x)  pthread_mutex_destroy (x)
+# define mutex_lock(x)     pthread_mutex_lock    (x)
+# define mutex_unlock(x)   pthread_mutex_unlock  (x)
+# define globally_lock()   _glob_lock ()
+# define globally_unlock() mutex_unlock (&glob_mutex)
+static void
+_glob_lock
+  (void)
+{
+  if (glob_mutex_initialized == 0) {
+    mutex_init (&glob_mutex);
+    glob_mutex_initialized = 1;
+  }
+  mutex_lock (&glob_mutex);
+}
+ #else
+# define mutex_init(x)     /* empty */
+# define mutex_lock(x)     /* empty */
+# define mutex_unlock(x)   /* empty */
+# define mutex_destroy(x)  /* empty */
+# define globally_lock()   /* empty */
+# define globally_unlock() /* empty */
+#endif
+
 /* ------------------------------------------------------------------------- *
  *                      private functions                                    *
  * ------------------------------------------------------------------------- */
@@ -141,9 +189,39 @@ find_bucket_ptr
       return Q ;
     Q = &q->next ;
   }
+  errno = ENOENT ;
   return 0;
 }
 
+/* qsort call back functions */
+static int 
+__comp
+  (hashqueue**  left, 
+   hashqueue** right)
+{
+  int n, min ;
+  
+  if ((min = (*left)->keylen) > (*right)->keylen)
+    min = (*right)->keylen ;
+  
+  if ((n = memcmp ((*left)->key, (*right)->key, min)) != 0)
+    return n;
+  
+  return (*left)->keylen - (*right)->keylen ;
+}
+
+static int 
+__comp_custom
+  (hashqueue**  left, 
+   hashqueue** right)
+{
+  return (*sorter_fn) (sorter_desc,
+		       (*left)->key,
+		       (*left)->keylen,
+		       (*right)->key,
+		       (*right)->keylen);
+}
+
 /* ------------------------------------------------------------------------- *
  *                 public functions: open/close management                   *
  * ------------------------------------------------------------------------- */
@@ -159,6 +237,10 @@ create_hlst
 
   if (estimated_size_hint == 0)
     estimated_size_hint = DEFAULT_BUCKETS ;
+  
+  /* adjust accoording to policy */
+  estimated_size_hint *= size_hint_percentage_compressor ;
+  estimated_size_hint /= 100 ;
 
   /* find appropriate list size, will stop at the last entry */
 # ifdef _WIN32
@@ -196,27 +278,36 @@ copy_hlst
   unsigned i, copy_only ;
   
   /* sanity check */
-  if (h == 0) return 0 ;
-
+  if (h == 0) {
+    errno = EINVAL;
+    return 0 ;
+  }
   if (estimated_size_hint == 0)
     /* get default from list to copy */
     hd = &h->z ;
 
-  else if (estimated_size_hint != h->z.mod)
-    /* find appropriate list size, will stop at the last entry */
-# ifdef _WIN32
-    for (;;) {
-      const hash_defs *hd1 = hd+1 ;
-      if (hd1->mod == 0 || hd1->mod > estimated_size_hint)
-        break ;
-      ++ hd ;
-    }
-# else
-    while (hd [1].mod != 0 && hd [1].mod <= estimated_size_hint)
-      ++ hd ;
-# endif
+  else {
+    /* adjust accoording to policy */
+    estimated_size_hint *= size_hint_percentage_compressor ;
+    estimated_size_hint /= 100 ;
 
-  new = (copy_only = (hd->mod == h->z.mod)) == 0
+    if (estimated_size_hint != h->z.mod) {
+      /* find appropriate list size, will stop at the last entry */
+#     ifdef _WIN32
+      for (;;) {
+	const hash_defs *hd1 = hd+1 ;
+	if (hd1->mod == 0 || hd1->mod > estimated_size_hint)
+	  break ;
+	++ hd ;
+      }
+#     else
+      while (hd [1].mod != 0 && hd [1].mod <= estimated_size_hint)
+	++ hd ;
+#     endif
+    }
+  }
+
+  new = (copy_only = (hd->mod == h->z.mod && copy == 0)) == 0
 
     /* create a new list */
     ? XMALLOC (sizeof (hlst) + (hd->mod - 1) * sizeof (void*))
@@ -225,9 +316,10 @@ copy_hlst
     : XMCOPY (h, sizeof (hlst) + (h->z.mod - 1) * sizeof (void*))
     ;
   
-  new->walk       =     0 ;
-  new->clup       =  clup ;
-  new->clup_state = state ;
+  new->walk          =                0 ;
+  new->clup          =             clup ;
+  new->clup_state    =            state ;
+  new->total_entries = h->total_entries ;
   
   /* we organize the new list while looping over the old one */
   for (i = 0; i < h->z.mod; i ++) {
@@ -257,14 +349,23 @@ copy_hlst
 
 	/* create new entry */
 	if ((Q = make_hlst ((hlst*)new, p->key, p->keylen)) == 0) {
-	  fputs ( __FILE__ "[copy_hlist ()]: corrupt hash list - aborting",
-		  stderr);  /* cannot happen, can it ? */
+	  fprintf (stderr, __FILE__ 
+		   "(%d): [make_hlst() == 0] serious bug, "
+		   "corrupt target list -- please report, aborting.\n",
+		   __LINE__);
 	  exit (2) ;
 	}
       }
 	
-      if (copy != 0) /* duplicate user contents */
-	*Q = (*copy) (cpstate, p->contents, p->key, p->keylen) ;
+      if (copy != 0 && /* duplicate user contents */
+	  (*Q = (*copy) (cpstate, p->contents, p->key, p->keylen)) == 0 &&
+	  errno != 0) {
+	int e = errno ;
+	/* stop: clean up */
+	destroy_hlst (new);
+	errno = e ;
+	return 0;
+      }
       
       /* get next template */
       p = p->next ;
@@ -291,23 +392,21 @@ flush_hlst
     clup = h->clup ;
     desc = h->clup_state ;
   }
-
   /* remove sorter */
   if (h->access != 0) {
     XFREE (h->access);
     h->access = 0;
   }
   for (i = 0; i < h->z.mod; i ++) {
-    hashqueue *p = h->bucket [i] ;
-    while (p != 0) {
-      hashqueue *q = p->next ;
+    /* do with this bucket */
+    hashqueue *p, **P = (hashqueue**)h->bucket + i ;
+    while ((p = *P) != 0) {
+      /* unlink that node, so even circular sublists would not loop */
+      *P = p->next ;
       if (clup != 0 && p->contents != 0)
 	(*clup) (desc, p->contents, p->key, p->keylen);
       XFREE (p);
-      p = q ;
     }
-    /* done with this bucket */
-    h->bucket [i] = 0 ;
   } /* for */
 
   /* cannot visit any node, anymore */
@@ -320,6 +419,8 @@ flush_hlst
     }
 #   endif
   }
+  /* statistics */
+  h->total_entries = 0 ;
 }
 
 
@@ -350,13 +451,18 @@ find_hlst
   int inx ;
 
   /* sanity check */
-  if (h == 0 || key == 0) return 0;
+  if (h == 0 || key == 0) {
+    errno = EINVAL;
+    return 0;
+  }
   
   GET_HASH_AND_LEN (h, inx, len, key);
 
-  return (Q = find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len)) == 0
-    ? 0 
-    : &(*Q)->contents ;
+  if ((Q = find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len)) != 0)
+    return &(*Q)->contents ;
+
+  errno = ENOENT;
+  return 0;
 }
 
 
@@ -371,13 +477,18 @@ make_hlst
   int inx ;
 
   /* sanity check */
-  if (h == 0 || key == 0) return 0;
+  if (h == 0 || key == 0) {
+    errno = EINVAL;
+    return 0;
+  }
   
   GET_HASH_AND_LEN (h, inx, len, key);
 
   /* cannot have duplicate entries */
-  if (find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len) != 0)
+  if (find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len) != 0) {
+    errno = EEXIST;
     return 0;
+  }
 
   /* make entry */
   q = XMALLOC (sizeof (hashqueue) + len - 1);
@@ -410,12 +521,17 @@ delete_hlst
   unsigned inx ;
 
   /* sanity check */
-  if (h == 0 || key == 0) return 0;
+  if (h == 0 || key == 0) {
+    errno = EINVAL;
+    return -1;
+  }
   
   GET_HASH_AND_LEN (h, inx, len, key);
 
-  if ((Q = find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len)) == 0)
+  if ((Q = find_bucket_ptr ((hashqueue**)h->bucket + inx, key, len)) == 0) {
+    errno = ENOENT ;
     return -1 ;
+  }
 
   q = *Q ;
 
@@ -449,24 +565,37 @@ delete_hlst
  *                 public functions: misc                                    *
  * ------------------------------------------------------------------------- */
 
+/* global factor to be applied internally to any estimated_size_hint */
+unsigned 
+compress_hlst_index 
+  (unsigned size_hint_percentage)
+{
+  unsigned o_index = size_hint_percentage_compressor ;
+  if ((size_hint_percentage_compressor = size_hint_percentage) > 100)
+    size_hint_percentage_compressor = 100 ;
+  return o_index;
+}
+
 char *
 query_key_hlst
   (void **t)
 {
-  return t == 0 
-    ? 0
-    : REVERT_FIELD_PTR (t, hashqueue, contents)->key 
-    ;
+  if (t == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  return REVERT_FIELD_PTR (t, hashqueue, contents)->key ;
 }
 
 unsigned
 query_keylen_hlst
   (void **t)
 {
-  return t == 0 
-    ? 0
-    : REVERT_FIELD_PTR (t, hashqueue, contents)->keylen 
-    ;
+  if (t == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  return  REVERT_FIELD_PTR (t, hashqueue, contents)->keylen ;
 }
 
 
@@ -475,10 +604,12 @@ int
 query_tranum_hlst
   (void **t)
 {
-  return t == 0 
-    ? 0
-    : REVERT_FIELD_PTR (t, hashqueue, contents)->tranum
-    ;
+  if (t == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  errno = 0 ;
+  return REVERT_FIELD_PTR (t, hashqueue, contents)->tranum ;
 }
 
 int
@@ -487,43 +618,27 @@ set_tranum_hlst
    int    n)
 {
   int last ;
-  if (t == 0) return 0;
+  if (t == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  errno = 0 ;
   last = REVERT_FIELD_PTR (t, hashqueue, contents)->tranum ;
   REVERT_FIELD_PTR (t, hashqueue, contents)->tranum = n ;
   return last;
 }
-
-
-int
-query_tranum_cont
-  (hlst *h)
-{
-  return h == 0 
-    ? 0
-    : h->tra_cnt
-    ;
-}
-
-
-int
-next_tranum_count
-  (hlst *h)
-{
-  return h == 0 
-    ? 0
-    : (++ h->tra_cnt < 0 ? (h->tra_cnt = 1) : h->tra_cnt)
-    ;
-}
 #endif /* ENABLE_RHLST */
 
 unsigned
 query_hlst_size
   (hlst *h)
 {
-  return h == 0 
-    ? 0
-    : h->total_entries
-    ;
+  if (h == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  errno = 0 ;
+  return h->total_entries ;
 }
 
 
@@ -538,7 +653,10 @@ open_hlst_search
   hsrch *s ;
 
   /* sanity check */
-  if (h == 0) return 0;
+  if (h == 0) {
+    errno = EINVAL;
+    return 0;
+  }
   
   s = XMALLOC (sizeof (hsrch));
   s->hlist     =       h ;	/* current hash list, to walk on */
@@ -558,20 +676,25 @@ next_hlst_search
   void ** V ;
 
   /* sanity check */
-  if (s == 0) return 0;
-
-  /* get the hash list */
-  if ((h = s->hlist) == 0)
-    /* list has been flushed */
+  if (s == 0) {
+    errno = EINVAL;
     return 0;
-
+  }
+  /* get the hash list */
+  if ((h = s->hlist) == 0) {
+    /* list has been flushed */
+    errno = ENOENT ;
+    return 0;
+  }
   if (s->ntry != 0)
     s->ntry->locked -- ;	/* release that node */
   else
     do {			/* find next node */
       /* get bucket, check for end-of-list */
-      if (++ s->bucket_id >= h->z.mod)
+      if (++ s->bucket_id >= h->z.mod) {
+	errno = 0;
 	return 0 ;
+      }
       s->ntry = h->bucket [s->bucket_id] ;
     } while (s->ntry == 0) ;
   
@@ -615,7 +738,7 @@ close_hlst_search
     }
     if (u->next == u) { /* XXXXXXXXXXXXX Grrr - should not happen (jh) */
       fprintf (stderr, 
-	       "%s (%n): [u->next == u] serious bug -- please report\n",
+	       "%s (%d): [u->next == u] serious bug -- please report\n",
 	       __FILE__,
 	       __LINE__);
       u->next = 0 ;
@@ -636,8 +759,10 @@ for_hlst_do
   int n ;
   
   /* sanity check */
-  if (h == 0 || fn == 0) return -1;
-
+  if (h == 0 || fn == 0) {
+    errno = EINVAL;
+    return -1;
+  }
   /* looping over the buckets */
   for (i = 0; i < h->z.mod; i ++) {
     
@@ -647,7 +772,9 @@ for_hlst_do
     while (p != 0) {
       /* get next, the cb function could delete the current entry */
       hashqueue *q = p->next ;
-      if ((n = (*fn) (state, p->contents, p->key, p->keylen)) != 0)
+      if ((n = (*fn) (state, p->contents, p->key, p->keylen)) < 0)
+	return -1;
+      if (n)
 	return n;
       p = q ;
     }
@@ -660,29 +787,13 @@ for_hlst_do
  *                 public functions: sorter index                            *
  * ------------------------------------------------------------------------- */
 
-static int 
-__comp
-  (hashqueue**  left, 
-   hashqueue** right)
-{
-  int n, min ;
-  
-  if ((min = (*left)->keylen) > (*right)->keylen)
-    min = (*right)->keylen ;
-  
-  if ((n = memcmp ((*left)->key, (*right)->key, min)) != 0)
-     return n;
-  
-  return (*left)->keylen - (*right)->keylen ;
-}
-
-
 void
 sort_hlst
   (hlst *h)
 {
   unsigned i;
   hashqueue **ix ;
+  int (*sorter_cb)(const void*,const void*);
 
   if (h == 0) return ;
 
@@ -712,26 +823,59 @@ sort_hlst
       p = p->next ;
     }
   }
-  
+
+  /* check comparison function */
+  if (h->sorter_fn != 0) {
+    globally_lock () ; /* not thread safe, otherwise */
+    sorter_fn   = h->sorter_fn   ;
+    sorter_desc = h->sorter_desc ;
+    sorter_cb   = (int(*)(const void*, const void*))__comp_custom ;
+  } else {
+    sorter_cb   = (int(*)(const void*, const void*))__comp ;
+  }
   /* sort that access array */
-  qsort (h->access->inx, h->total_entries, sizeof (hashqueue*),
-	 (int(*)(const void*, const void*))__comp);
+  qsort (h->access->inx, h->total_entries, sizeof (hashqueue*), sorter_cb);
+  
+  if (h->sorter_fn != 0)
+  {
+    globally_unlock () ;
+  }
 }
 
 
+int
+csort_hlst
+  (hlst *h,
+   int (*fn)(void*,const char*,unsigned,const char*,unsigned),
+   void *fn_desc)
+{
+  if (h == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  h->sorter_fn   = fn;
+  h->sorter_desc = fn_desc;
+  return 0;
+}
+
 void **
 inx_hlst
   (hlst    *h,
    unsigned n)
 {
   hashqueue *p ;
-  
-  if (h != 0 &&
-      h->access != 0 &&
-      n < h->access->size &&
-      (p = h->access->inx [n]) != 0)
-    return &p->contents ;
 
+  if (h == 0) {
+    errno = EINVAL;
+    return 0;
+  }
+  if (h->access == 0) {
+    errno = ESRCH;
+    return 0;
+  }
+  if (n < h->access->size && (p = h->access->inx [n]) != 0)
+    return &p->contents ;
+  errno = ENOENT;
   return 0 ;
 }
 
@@ -751,9 +895,13 @@ unsort_hlst
 
 int
 hlst_buckets
-(hlst *h)
+  (hlst *h)
 {
-  return h == 0 ? -1 : h->z.mod;
+  if (h == 0) {
+    errno = EINVAL;
+    return -1;
+  }
+  return h->z.mod;
 }
 
 
