@@ -2282,7 +2282,7 @@ int internal_recv(int soc, char ** data, int * data_sz, int * msg_type )
  if ( len >= sz )
  {
   sz = len + 1;
-  buf = erealloc( buf, len );
+  buf = erealloc( buf, sz );
  }
 
  if ( len > 0 )
