@@ -1294,7 +1294,9 @@ void _add_plugin_preference(prefs, p_name, name, type, defaul)
 
  pref = emalloc(strlen(p_name)+10+strlen(type)+strlen(cname));
  sprintf(pref, "%s[%s]:%s", p_name, type, cname);
- arg_add_value(prefs, pref, ARG_STRING, strlen(defaul), estrdup(defaul));
+ if ( arg_get_value(prefs, pref) == NULL )
+  arg_add_value(prefs, pref, ARG_STRING, strlen(defaul), estrdup(defaul));
+
  efree(&cname);
  efree(&pref);
 }
@@ -1989,15 +1991,20 @@ find_in_path(name, safe)
       sprintf(p2, "/%s", name);
       if (access(cmd, X_OK) == 0)
 	{
-	  *p2 = '\0';
+	  struct stat	st;
+	  if (stat(cmd, &st) < 0)
+	    perror(cmd);
+	  else if (S_ISREG(st.st_mode))
+	    {
+	      *p2 = '\0';
 #if 0
-	  fprintf(stderr, "find_in_path: %s found in %s\n", name, cmd);
+	      fprintf(stderr, "find_in_path: %s found in %s\n", name, cmd);
 #endif
-	  return cmd;
+	      return cmd;
+	    }
 	}
 #if 0
-      else
-	fprintf(stderr, "find_in_path: No %s\n", cmd);
+	  fprintf(stderr, "find_in_path: No %s\n", cmd);
 #endif
     }
   return NULL;
